/*********************************************************************************/
/*  Copyright 2015 Barcelona Supercomputing Center                               */
/*                                                                               */
/*  This file is part of the DLB library.                                        */
/*                                                                               */
/*  DLB is free software: you can redistribute it and/or modify                  */
/*  it under the terms of the GNU Lesser General Public License as published by  */
/*  the Free Software Foundation, either version 3 of the License, or            */
/*  (at your option) any later version.                                          */
/*                                                                               */
/*  DLB is distributed in the hope that it will be useful,                       */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/*  GNU Lesser General Public License for more details.                          */
/*                                                                               */
/*  You should have received a copy of the GNU Lesser General Public License     */
/*  along with DLB.  If not, see <http://www.gnu.org/licenses/>.                 */
/*********************************************************************************/

#define _GNU_SOURCE
#include <sched.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>

#include "LB_comm/shmem.h"
#include "LB_comm/shmem_stats.h"
#include "support/tracing.h"
#include "support/debug.h"
#include "support/options.h"
#include "support/mask_utils.h"

/* NOTE on default values:
 * The shared memory will be initializated to 0 when created,
 * thus all default values should represent 0
 * A CPU, by default, is DISABLED and owned by NOBODY
 */

#define NOBODY 0
#define ME getpid()

typedef pid_t spid_t;  // Sub-process ID

typedef enum {
    DISABLED = 0,        // Not owned by any process nor part of the DLB mask
    BUSY,
    LENT
} cpu_state_t;

typedef struct {
    spid_t      owner;   // Original owner, useful to resolve reclaimed CPUs
    spid_t      guest;   // Current user of the CPU
    cpu_state_t state;
} cpu_info_t;

typedef struct {
    cpu_info_t node_info[0];
} shdata_t;

static shmem_handler_t *shm_handler;
static shdata_t *shdata;
static cpu_set_t default_mask;   // default mask of the process
static cpu_set_t affinity_mask;  // affinity mask of the process
static cpu_set_t dlb_mask;       // CPUs not owned by any process but usable by others
static int cpus_node;
static bool cpu_is_public_post_mortem = false;

static inline bool is_idle( int cpu ) {
    return (shdata->node_info[cpu].state == LENT) && (shdata->node_info[cpu].guest == NOBODY);
}

void shmem_cpuarray__init( const cpu_set_t *cpu_set ) {
    mu_parse_mask( options_get_mask(), &dlb_mask );
    memcpy( &default_mask, cpu_set, sizeof(cpu_set_t) );
    mu_get_affinity_mask( &affinity_mask, &default_mask, MU_ANY_BIT );
    cpus_node = mu_get_system_size();

    // Basic size + zero-length array real length
    shm_handler = shmem_init( (void**)&shdata, sizeof(shdata_t) + sizeof(cpu_info_t)*cpus_node, "lewi" );

    DLB_INSTR( int idle_count = 0; )

    shmem_lock( shm_handler );
    {
        int cpu;
        // Check first that my default_mask is not already owned
        for ( cpu = 0; cpu < cpus_node; cpu++ )
            if ( CPU_ISSET( cpu, &default_mask ) && shdata->node_info[cpu].owner != NOBODY ) {
                shmem_unlock( shm_handler );
                fatal0( "Another process in the same node is using one of your cpus" );
            }

        for ( cpu = 0; cpu < cpus_node; cpu++ ) {
            // Add my mask info
            if ( CPU_ISSET( cpu, &default_mask ) ) {
                shdata->node_info[cpu].owner = ME;
                shdata->node_info[cpu].state = BUSY;
                // My CPU could already have a guest if the DLB_mask was being used
                if ( shdata->node_info[cpu].guest == NOBODY ) {
                    shdata->node_info[cpu].guest = ME;
                    shmem_stats__update_cpu(cpu, STATS_OWNED);
                }
            }
            // Add DLB mask info
            if ( CPU_ISSET( cpu, &dlb_mask )
                    && shdata->node_info[cpu].owner == NOBODY
                    && shdata->node_info[cpu].state == DISABLED ) {
                shdata->node_info[cpu].state = LENT;
            }

            // Look for Idle CPUs
            DLB_INSTR( if (is_idle(cpu)) idle_count++; )
        }
    }
    shmem_unlock( shm_handler );

    verbose( VB_SHMEM, "Default Mask: %s", mu_to_str(&default_mask) );
    verbose( VB_SHMEM, "Default Affinity Mask: %s", mu_to_str(&affinity_mask) );

    add_event( IDLE_CPUS_EVENT, idle_count );
}

void shmem_cpuarray__finalize( void ) {
    DLB_INSTR( int idle_count = 0; )

    shmem_lock( shm_handler );
    {
        int cpu;
        for ( cpu = 0; cpu < cpus_node; cpu++ ) {
            if ( CPU_ISSET( cpu, &default_mask ) ) {
                shdata->node_info[cpu].owner = NOBODY;
                if ( shdata->node_info[cpu].guest == ME ) {
                    shdata->node_info[cpu].guest = NOBODY;
                    shmem_stats__update_cpu(cpu, STATS_IDLE);
                }
                if ( cpu_is_public_post_mortem
                        || CPU_ISSET( cpu, &dlb_mask ) ) {
                    shdata->node_info[cpu].state = LENT;
                } else {
                    shdata->node_info[cpu].state = DISABLED;
                }
            } else {
                // Free external CPUs that may I be using
                if ( shdata->node_info[cpu].guest == ME ) {
                    shdata->node_info[cpu].guest = NOBODY;
                    shmem_stats__update_cpu(cpu, STATS_IDLE);
                }
            }

            if ( is_idle(cpu) ) {
                DLB_INSTR( idle_count++; )
            }
        }
    }
    shmem_unlock( shm_handler );

    add_event( IDLE_CPUS_EVENT, idle_count );

    shmem_finalize( shm_handler );
}

/* Add cpu_mask to the Shared Mask
 * If the process originally owns the CPU:      State => LENT
 * If the process is currently using the CPU:   Guest => NOBODY
 */
void shmem_cpuarray__add_mask( const cpu_set_t *cpu_mask ) {
    DLB_DEBUG( cpu_set_t freed_cpus; )
    DLB_DEBUG( cpu_set_t idle_cpus; )
    DLB_DEBUG( CPU_ZERO( &freed_cpus ); )
    DLB_DEBUG( CPU_ZERO( &idle_cpus ); )

    DLB_INSTR( int idle_count = 0; )

    shmem_lock( shm_handler );
    {
        int cpu;
        for ( cpu = 0; cpu < cpus_node; cpu++ ) {
            if ( CPU_ISSET( cpu, cpu_mask ) ) {

                // If the CPU was mine, just change the state
                if ( CPU_ISSET( cpu, &default_mask ) ) {
                    shdata->node_info[cpu].state = LENT;
                }

                // If am currently using the CPU, free it
                if ( shdata->node_info[cpu].guest == ME ) {
                    shdata->node_info[cpu].guest = NOBODY;
                    shmem_stats__update_cpu(cpu, STATS_IDLE);
                    DLB_DEBUG( CPU_SET( cpu, &freed_cpus ); )
                }
            }

            // Look for Idle CPUs, only in DEBUG or INSTRUMENTATION
            if ( is_idle(cpu) ) {
                DLB_INSTR( idle_count++; )
                DLB_DEBUG( CPU_SET( cpu, &idle_cpus ); )
            }
        }
    }
    shmem_unlock( shm_handler );

    DLB_DEBUG( int size = CPU_COUNT( &freed_cpus ); )
    DLB_DEBUG( int post_size = CPU_COUNT( &idle_cpus); )
    verbose( VB_SHMEM, "Lending %s", mu_to_str(&freed_cpus) );
    verbose( VB_SHMEM, "Increasing %d Idle Threads (%d now)", size, post_size );
    verbose( VB_SHMEM, "Available mask: %s", mu_to_str(&idle_cpus) );

    add_event( IDLE_CPUS_EVENT, idle_count );
}

/* Add cpu to the Shared Mask
 * If the process originally owns the CPU:      State => LENT
 * If the process is currently using the CPU:   Guest => NOBODY
 */
void shmem_cpuarray__add_cpu( int cpu ) {
    DLB_DEBUG( cpu_set_t freed_cpus; )
    DLB_DEBUG( cpu_set_t idle_cpus; )
    DLB_DEBUG( CPU_ZERO( &freed_cpus ); )
    DLB_DEBUG( CPU_ZERO( &idle_cpus ); )

    DLB_INSTR( int idle_count = 0; )

    shmem_lock( shm_handler );
    {
        // If the CPU was mine, just change the state
        if ( CPU_ISSET( cpu, &default_mask ) ) {
            shdata->node_info[cpu].state = LENT;
        }

        // If am currently using the CPU, free it
        if ( shdata->node_info[cpu].guest == ME ) {
            shdata->node_info[cpu].guest = NOBODY;
            shmem_stats__update_cpu(cpu, STATS_IDLE);
            DLB_DEBUG( CPU_SET( cpu, &freed_cpus ); )
        }

        // Look for Idle CPUs, only in DEBUG or INSTRUMENTATION
        int i;
        for ( i = 0; i < cpus_node; i++ ) {
            if ( is_idle(i) ) {
                DLB_INSTR( idle_count++; )
                DLB_DEBUG( CPU_SET( i, &idle_cpus ); )
            }
        }
    }
    shmem_unlock( shm_handler );

    DLB_DEBUG( int size = CPU_COUNT( &freed_cpus ); )
    DLB_DEBUG( int post_size = CPU_COUNT( &idle_cpus); )
    verbose( VB_SHMEM, "Lending %s", mu_to_str(&freed_cpus) );
    verbose( VB_SHMEM, "Increasing %d Idle Threads (%d now)", size, post_size );
    verbose( VB_SHMEM, "Available mask: %s", mu_to_str(&idle_cpus) );

    add_event( IDLE_CPUS_EVENT, idle_count );
}

/* Remove the process default mask from the Shared Mask
 * CPUs from default_mask:          State => BUSY
 * CPUs that also have no guest:    Guest => ME
 */
const cpu_set_t* shmem_cpuarray__recover_defmask( void ) {
    DLB_DEBUG( cpu_set_t recovered_cpus; )
    DLB_DEBUG( cpu_set_t idle_cpus; )
    DLB_DEBUG( CPU_ZERO( &recovered_cpus ); )
    DLB_DEBUG( CPU_ZERO( &idle_cpus ); )

    DLB_INSTR( int idle_count = 0; )

    shmem_lock( shm_handler );
    {
        int cpu;
        for ( cpu = 0; cpu < cpus_node; cpu++ ) {
            if ( CPU_ISSET( cpu, &default_mask ) ) {
                shdata->node_info[cpu].state = BUSY;
                if ( shdata->node_info[cpu].guest == NOBODY ) {
                    shdata->node_info[cpu].guest = ME;
                    shmem_stats__update_cpu(cpu, STATS_OWNED);
                    DLB_DEBUG( CPU_SET( cpu, &recovered_cpus ); )
                }
            }

            // Look for Idle CPUs, only in DEBUG or INSTRUMENTATION
            if ( is_idle(cpu) ) {
                DLB_INSTR( idle_count++; )
                DLB_DEBUG( CPU_SET( cpu, &idle_cpus ); )
            }
        }
    }
    shmem_unlock( shm_handler );

    DLB_DEBUG( int recovered = CPU_COUNT( &recovered_cpus); )
    DLB_DEBUG( int post_size = CPU_COUNT( &idle_cpus); )
    verbose( VB_SHMEM, "Decreasing %d Idle Threads (%d now)", recovered, post_size );
    verbose( VB_SHMEM, "Available mask: %s", mu_to_str(&idle_cpus) );

    add_event( IDLE_CPUS_EVENT, idle_count );

    return &default_mask;
}

/* Remove part of the process default mask from the Shared Mask
 * CPUs from default_mask:          State => BUSY
 * CPUs that also have no guest:    Guest => ME
 */
void shmem_cpuarray__recover_some_defcpus( cpu_set_t *mask, int max_resources ) {
    DLB_DEBUG( cpu_set_t recovered_cpus; )
    DLB_DEBUG( cpu_set_t idle_cpus; )
    DLB_DEBUG( CPU_ZERO( &recovered_cpus ); )
    DLB_DEBUG( CPU_ZERO( &idle_cpus ); )

    DLB_INSTR( int idle_count = 0; )

    shmem_lock( shm_handler );
    {
        int cpu;
        for ( cpu = 0; (cpu < cpus_node) && (max_resources > 0); cpu++ ) {
            if ( (CPU_ISSET(cpu, &default_mask)) && (!CPU_ISSET(cpu, mask)) ) {
                shdata->node_info[cpu].state = BUSY;
                CPU_SET( cpu, mask );
                max_resources--;
                if ( shdata->node_info[cpu].guest == NOBODY ) {
                    shdata->node_info[cpu].guest = ME;
                    shmem_stats__update_cpu(cpu, STATS_OWNED);
                    DLB_DEBUG( CPU_SET( cpu, &recovered_cpus ); )
                }
            }

            // Look for Idle CPUs, only in DEBUG or INSTRUMENTATION
            if ( is_idle(cpu) ) {
                DLB_INSTR( idle_count++; )
                DLB_DEBUG( CPU_SET( cpu, &idle_cpus ); )
            }
        }
    }
    shmem_unlock( shm_handler );

    DLB_DEBUG( int recovered = CPU_COUNT( &recovered_cpus); )
    DLB_DEBUG( int post_size = CPU_COUNT( &idle_cpus); )
    verbose( VB_SHMEM, "Decreasing %d Idle Threads (%d now)", recovered, post_size );
    verbose( VB_SHMEM, "Available mask: %s", mu_to_str(&idle_cpus) );

    add_event( IDLE_CPUS_EVENT, idle_count );
}

/* Remove CPU from the Shared Mask
 * CPUs from default_mask:          State => BUSY
 * CPUs that also have no guest:    Guest => ME
 */
void shmem_cpuarray__recover_cpu( int cpu ) {
    DLB_DEBUG( cpu_set_t recovered_cpus; )
    DLB_DEBUG( cpu_set_t idle_cpus; )
    DLB_DEBUG( CPU_ZERO( &recovered_cpus ); )
    DLB_DEBUG( CPU_ZERO( &idle_cpus ); )

    DLB_INSTR( int idle_count = 0; )

    shmem_lock( shm_handler );
    {
        if ( CPU_ISSET(cpu, &default_mask) ) {
            shdata->node_info[cpu].state = BUSY;
            if ( shdata->node_info[cpu].guest == NOBODY ) {
                shdata->node_info[cpu].guest = ME;
                shmem_stats__update_cpu(cpu, STATS_OWNED);
                DLB_DEBUG( CPU_SET( cpu, &recovered_cpus ); )
            }
        }

        // Look for Idle CPUs, only in DEBUG or INSTRUMENTATION
        if ( is_idle(cpu) ) {
            DLB_INSTR( idle_count++; )
                DLB_DEBUG( CPU_SET( cpu, &idle_cpus ); )
        }
    }
    shmem_unlock( shm_handler );

    DLB_DEBUG( int recovered = CPU_COUNT( &recovered_cpus); )
    DLB_DEBUG( int post_size = CPU_COUNT( &idle_cpus); )
    verbose( VB_SHMEM, "Decreasing %d Idle Threads (%d now)", recovered, post_size );
    verbose( VB_SHMEM, "Available mask: %s", mu_to_str(&idle_cpus) );

    add_event( IDLE_CPUS_EVENT, idle_count );
}

/* Remove non default CPUs that have been set as BUSY by their owner
 * or remove also CPUs that habe been set DISABLED
 * Reclaimed CPUs:    Guest => Owner
 */
int shmem_cpuarray__return_claimed ( cpu_set_t *mask ) {
    int returned = 0;

    DLB_DEBUG( cpu_set_t idle_cpus; )
    DLB_DEBUG( cpu_set_t returned_cpus; )
    DLB_DEBUG( CPU_ZERO( &idle_cpus ); )
    DLB_DEBUG( CPU_ZERO( &returned_cpus ); )

    DLB_INSTR( int idle_count = 0; )

    shmem_lock( shm_handler );
    {
        int cpu;
        for ( cpu = 0; cpu < cpus_node; cpu++ ) {
            if ( CPU_ISSET( cpu, mask ) && (shdata->node_info[cpu].owner != ME) && shdata->node_info[cpu].state!=LENT ) {
                if ( shdata->node_info[cpu].guest == ME ) {
                    shdata->node_info[cpu].guest = shdata->node_info[cpu].owner;
                    shmem_stats__update_cpu(cpu, STATS_OWNED);
                }
                returned++;
                CPU_CLR( cpu, mask );
                DLB_DEBUG( CPU_SET( cpu, &returned_cpus ); )
            }

            // Look for Idle CPUs, only in DEBUG or INSTRUMENTATION
            if ( is_idle(cpu) ) {
                DLB_INSTR( idle_count++; )
                DLB_DEBUG( CPU_SET( cpu, &idle_cpus ); )
            }
        }
    }
    shmem_unlock( shm_handler );
    if ( returned > 0 ) {
        verbose( VB_SHMEM, "Giving back %d Threads (%s)", returned, mu_to_str(&returned_cpus) );
        verbose( VB_SHMEM, "Available mask: %s", mu_to_str(&idle_cpus) );
    }

    add_event( IDLE_CPUS_EVENT, idle_count );

    return returned;
}

/* Collect as many idle CPUs as indicated by max_resources
 * Affine CPUs have preference
 * Idle CPUs are the ones that (state == LENT) && (guest == NOBODY)
 * If successful:       Guest => ME
 */
int shmem_cpuarray__collect_mask ( cpu_set_t *mask, int max_resources ) {
    int cpu;
    int collected1 = 0;
    int collected2 = 0;
    bool some_idle_cpu = false;

    DLB_DEBUG( cpu_set_t idle_cpus; )
    DLB_DEBUG( CPU_ZERO( &idle_cpus ); )

    DLB_INSTR( int idle_count = 0; )

    // Fast non-safe check to get if there is some idle CPU
    for ( cpu = 0; cpu < cpus_node; cpu++ ) {
        some_idle_cpu = is_idle( cpu );
        if ( some_idle_cpu ) { break; }
    }

    if ( some_idle_cpu && max_resources > 0) {

        shmem_lock( shm_handler );
        {
            /* First Step: Retrieve affine cpus */
            for ( cpu = 0; (cpu < cpus_node) && (max_resources > 0); cpu++ ) {
                if ( CPU_ISSET( cpu, &affinity_mask )
                        && shdata->node_info[cpu].state == LENT
                        && shdata->node_info[cpu].guest == NOBODY ) {
                    shdata->node_info[cpu].guest = ME;
                    shmem_stats__update_cpu(cpu, STATS_GUESTED);
                    CPU_SET( cpu, mask );
                    max_resources--;
                    collected1++;
                }
            }
            verbose( VB_SHMEM, "Getting %d affine Threads (%s)", collected1, mu_to_str(mask) );

//            /* Second Step: Retrieve non-affine cpus, if needed */
//            cpu_set_t free_cpus; //mask with free cpus
//            cpu_set_t free_sockets; //mask with full free sockets
//            
//            //first fill the free cpus mask
//            for ( cpu = 0; cpu < cpus_node; cpu++ ) {   
//                if ( shdata->node_info[cpu].state == LENT
//                        && shdata->node_info[cpu].guest == NOBODY ) {
//                    CPU_SET(cpu, &free_cpus);
//                }
//            }
//            //Get the full sockets
//            mu_get_affinity_mask( &free_sockets, &free_cpus, MU_ALL_BITS );
//
//            for ( cpu = 0; (cpu < cpus_node) && (max_resources > 0); cpu++ ) {
//                if ( CPU_ISSET( cpu, &free_sockets) ) {
//                    shdata->node_info[cpu].guest = ME;
//                    CPU_SET( cpu, mask );
//                    max_resources--;
//                    collected2++;
//                }
//            }

            //Original code to retrieve non affine CPUs
            for ( cpu = 0; (cpu < cpus_node) && (max_resources > 0); cpu++ ) {
                if ( shdata->node_info[cpu].state == LENT
                        && shdata->node_info[cpu].guest == NOBODY ) {
                    shdata->node_info[cpu].guest = ME;
                    shmem_stats__update_cpu(cpu, STATS_GUESTED);
                    CPU_SET( cpu, mask );
                    max_resources--;
                    collected2++;
                }
            }
            verbose( VB_SHMEM, "Getting %d other Threads (%s)", collected2, mu_to_str(mask) );

            // FIXME: Another loop, efficiency?
            // Look for Idle CPUs, only in DEBUG or INSTRUMENTATION
            for ( cpu = 0; cpu < cpus_node; cpu++ ) {
                if ( is_idle(cpu) ) {
                    DLB_INSTR( idle_count++; )
                    DLB_DEBUG( CPU_SET( cpu, &idle_cpus ); )
                }
            }
        }
        shmem_unlock( shm_handler );
    }

    int collected = collected1 + collected2;
    if ( collected > 0 ) {
        DLB_DEBUG( int post_size = CPU_COUNT( &idle_cpus ); )
        verbose( VB_SHMEM, "Clearing %d Idle Threads (%d left)", collected, post_size );
        verbose( VB_SHMEM, "Available mask: %s", mu_to_str(&idle_cpus) );
        add_event( IDLE_CPUS_EVENT, idle_count );
    }
    return collected;
}

/* Return false if my CPU is being used by other process
 */
bool shmem_cpuarray__is_cpu_borrowed ( int cpu ) {
    // If the CPU is not mine, skip check
    if ( !(CPU_ISSET(cpu, &default_mask)) ) {
        return true;
    }

    // If the CPU is free, assign it to myself
    if ( shdata->node_info[cpu].guest == NOBODY ) {
        shmem_lock( shm_handler );
        if ( shdata->node_info[cpu].guest == NOBODY ) {
            shdata->node_info[cpu].guest = ME;
            shmem_stats__update_cpu(cpu, STATS_OWNED);
        }
        shmem_unlock( shm_handler );
    }

    return shdata->node_info[cpu].guest == ME;
}

/* Return true if the CPU is foreign and reclaimed
 */
bool shmem_cpuarray__is_cpu_claimed( int cpu ) {
    if ( shdata->node_info[cpu].state == DISABLED ) {
        return true;
    }

    return (shdata->node_info[cpu].owner != ME) && (shdata->node_info[cpu].state == BUSY);
}

/*Reset current mask to use my default cpus
  I.e: Release borrowed cpus and claim lend cpus
 */
int shmem_cpuarray__reset_default_cpus(cpu_set_t *mask){
    shmem_lock( shm_handler );
    int i;
    int n=0;
    for ( i = 0; i < mu_get_system_size(); i++ ) {
        //CPU is mine --> claim it and set in my mask
        if ( CPU_ISSET(i, &default_mask) ) {
            shdata->node_info[i].state = BUSY;
            if ( shdata->node_info[i].guest == NOBODY ) {
                shdata->node_info[i].guest = ME;
                shmem_stats__update_cpu(i, STATS_OWNED);
            }
            CPU_SET( i, mask);
            n++;
        //CPU is not mine and I have it --> release it and clear from my mask
        }else if(CPU_ISSET(i, mask)){
            if ( shdata->node_info[i].guest == ME ){
                shdata->node_info[i].guest = NOBODY;
                shmem_stats__update_cpu(i, STATS_IDLE);
            }
            CPU_CLR( i, mask);
        }
    }
    ensure( CPU_EQUAL(mask, &default_mask) , "Mask not correctly updated on reset" );
    shmem_unlock( shm_handler );
    return n;

}

/*
  Acquire this cpu, wether it was mine or not
  Can have a problem if the cpu was borrowed by somebody else
  If force take it from the owner
 */
bool shmem_cpuarray__acquire_cpu (int cpu, bool force){

    bool acquired=true;

    shmem_lock( shm_handler );
    //cpu is mine -> Claim it
    if (CPU_ISSET(cpu, &default_mask)){
        shdata->node_info[cpu].state = BUSY;
        // It is important to not modify the 'guest' field to remark that the CPU is claimed

    //cpu is not mine but is free -> take it
    }else if(shdata->node_info[cpu].state == LENT
            && shdata->node_info[cpu].guest == NOBODY){
        shdata->node_info[cpu].guest = ME;
        shmem_stats__update_cpu(cpu, STATS_GUESTED);

    //cpus is not mine and the owner has recover it
    }else if (force){
        //If force -> take it
        if( shdata->node_info[cpu].guest==shdata->node_info[cpu].owner ){
            shdata->node_info[cpu].guest = ME;
            shmem_stats__update_cpu(cpu, STATS_GUESTED);

        }else{
            //FIXME: An other process has borrowed it
            // or    the cpu is DISABLED in the system
            acquired=false;
        }
    }else{
        //No force -> nothing
        acquired=false;
    }

    //    add_event( IDLE_CPUS_EVENT, idle_count );
    shmem_unlock( shm_handler );
    return acquired;

}

/* Update CPU ownership according to the new process mask.
 * To avoid collisions, we only release the ownership if we still own it
 */
void shmem_cpuarray__update_ownership(const cpu_set_t* process_mask) {
    shmem_lock( shm_handler );
    int cpu;
    for ( cpu = 0; cpu < mu_get_system_size(); cpu++ ) {
        if ( CPU_ISSET(cpu, process_mask) ) {
            // CPU ownership should be mine
            if ( shdata->node_info[cpu].owner != ME ) {
                // Steal CPU
                CPU_SET(cpu, &default_mask);
                shdata->node_info[cpu].owner = ME;
                shdata->node_info[cpu].guest = ME;
                shdata->node_info[cpu].state = BUSY;
                verbose(VB_SHMEM, "Acquiring ownership of CPU %d", cpu);
            }
        } else if ( shdata->node_info[cpu].owner == ME ) {
            // Release CPU ownership
            CPU_CLR(cpu, &default_mask);
            shdata->node_info[cpu].owner = NOBODY;
            shdata->node_info[cpu].state = DISABLED;
            verbose(VB_SHMEM, "Releasing ownership of CPU %d", cpu);
        }
    }
    shmem_unlock( shm_handler );
}

/* This function is intended to be called from external processes only to consult the shdata
 * That's why we should initialize and finalize the shared memory
 */
void shmem_cpuarray__print_info( void ) {
    cpus_node = mu_get_system_size();

    // Basic size + zero-length array real length
    shmem_handler_t *handler;
    handler = shmem_init( (void**)&shdata, sizeof(shdata_t) + sizeof(cpu_info_t)*cpus_node, "lewi" );
    cpu_info_t node_info_copy[cpus_node];

    shmem_lock( handler );
    {
        memcpy( node_info_copy, shdata->node_info, sizeof(cpu_info_t)*cpus_node );
    }
    shmem_unlock( handler );
    shmem_finalize( handler );

    char owners[512] = "OWNERS: ";
    char guests[512] = "GUESTS: ";
    char states[512] = "STATES: ";
    char *o = owners+8;
    char *g = guests+8;
    char *s = states+8;
    int cpu;
    for ( cpu=0; cpu<cpus_node; cpu++ ) {
        o += snprintf( o, 8, "%d, ", node_info_copy[cpu].owner );
        g += snprintf( g, 8, "%d, ", node_info_copy[cpu].guest );
        s += snprintf( s, 8, "%d, ", node_info_copy[cpu].state );
    }

    info0( owners );
    info0( guests );
    info0( states );
    info0( "States Legend: DISABLED=%d, BUSY=%d, LENT=%d", DISABLED, BUSY, LENT );
}

